#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "path.h"

static path_t *path_new(int sx, int sy, int dx, int dy);
static void path_resize(path_t *path);
static int path_compare(void *a, void *b);

static path_t *path_copy(path_t *path);
static void path_add_move(path_t *path, int move);

/* keep these up to date respect to MD_xxxx enums! */
static inline unsigned int dirtoindex(unsigned int dir)
{
	if (dir == MD_NORTH)
		return 0;
	else if (dir == MD_SOUTH)
		return 1;
	else if (dir == MD_EAST)
		return 2;
	else if (dir == MD_WEST)
		return 3;
	else 
		abort();
}


list_t *path_cache;
list_t *path_find_cache;

void init_path_engine(void)
{
	path_cache = list_new(path_compare);
	path_find_cache = list_new(path_compare);
}

static int path_priority_func(void *a, void *b)
{
	path_t *pa, *pb;
	pa = (path_t *) a;
	pb = (path_t *) b;
	return pa->goal_distance - pb->goal_distance;
}

static void path_find_initialize(map_t *map)
{
	size_t x, y;
	for (y = 1; y <= map->height; y++)
		for (x = 1; x <= map->width; x++)
			map_set_mindistance(map, x, y, ~0UL);
}

static path_t *path_find(map_t *map, int flags)
{
	unsigned int moves, move, mindistance;
	int i;
#ifdef PROFILE
	int num_iterations = 0;
#endif
	path_t *p, *p2;

	path_find_initialize(map);

	for (;;) {
		if ((p = list_car_delete(path_find_cache)) == NULL)
			break;

		if (p->cx == p->dx && p->cy == p->dy)
			break;	

		moves = possible_moves(map, p->cx, p->cy, 1);
		while (moves) {
			move = dirtoindex(moves & 0xff);
			p2 = path_copy(p);
			path_add_move(p2, move);


			p2->goal_distance += map_get_blacklist(map, p2->cx, p2->cy) * 10;
			p2->goal_distance += map_get_price(map, p2->cx, p2->cy);

//			p2->goal_distance += 
//				map_get_visited(map, p2->cx, p2->cy);

			mindistance = map_get_mindistance(map, p2->cx, p2->cy);
			if (mindistance <= p2->goal_distance) 
				path_free(p2);
			else {
				list_add_priority(path_find_cache, p2, path_priority_func);
				map_set_mindistance(map, p2->cx, p2->cy, p2->goal_distance);
			}
			moves >>= 8;
		}

		path_free(p);
	}

	return p;
}

path_t *path_compute(map_t *map, int sx, int sy, int dx, int dy, int flags)
{
	path_t *path, *npath;

	path = path_new(sx, sy, dx, dy);
	if ((npath = list_find(path_cache, path))) {
		free(path);
		return path_copy(npath);
	}

	while ((npath = list_car_delete(path_find_cache)))
		path_free(npath);

	list_add(path_find_cache, path);
	if ((npath = path_find(map, flags)) == NULL)
		return NULL;
	list_add(path_cache, npath);
	return path_copy(npath);
}

static path_t *path_new(int sx, int sy, int dx, int dy)
{
	path_t *path;
	path = (path_t *) malloc(sizeof(path_t));
	path->nsteps = 0;
	path->nalloc = 16;
	path->sx = path->cx = sx;
	path->sy = path->cy = sy;
	path->dx = dx;
	path->dy = dy;
	path->goal_distance = distance(sx, sy, dx, dy);
	path->moves = (uint8_t *) malloc(path->nalloc);
	return path;
}

static void path_add_move(path_t *path, int move)
{
	path_resize(path);
	path->moves[path->nsteps++] = move;
	path->cx += directions[move][0];
	path->cy += directions[move][1];
	path->goal_distance = distance(path->cx, path->cy, path->dx, 
		path->dy) + path->nsteps;
}

static path_t *path_copy(path_t *path)
{
	path_t *npath;

	npath = (path_t *) malloc(sizeof(path_t));
	memcpy(npath, path, sizeof(path_t));
	npath->moves = (uint8_t *) malloc(path->nalloc);
	memcpy(npath->moves, path->moves, path->nalloc);
	return npath;
}

void path_free(path_t *path)
{
	free(path->moves);
	free(path);
}

static void path_resize(path_t *path)
{
	if (path->nalloc > (path->nsteps + 1))
		return;
	path->moves = (uint8_t *) realloc(path->moves, path->nalloc *= 2);
}

static int path_compare(void *a, void *b)
{
	path_t *pa, *pb;
	pa = (path_t *) a;
	pb = (path_t *) b;
	return pa->sx == pb->sx && pa->sy == pb->sy &&
		pa->dx == pb->dx && pa->dy == pb->dy;
}
