#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <regex.h>

#include <ncurses.h>

#include "util.h"
#include "path.h"
#include "map.h"
#include "net.h"
#include "robot.h"
#include "package.h"

#define max(__a__, __b__) (((__a__) > (__b__)) ? (__a__) : (__b__))

static int iterations = 0;
static int wandercount = 0;

static intarray_t *packages_here = NULL;

void robot_move(robot_t *robot, map_t *map, int dx, int dy);
void compute_price_mask(map_t *map, robot_t *myrobot);

static intarray_t *packages = NULL;

robot_t *myrobot;
void purge_delivered_packages(void);

long int atoi2(char **p);

void segfault(int sig)
{
	logger("%d: segmentation fault\n", myrobot->id);
	signal(SIGSEGV, SIG_DFL);
}

static int init_game(sockbuf_t *sock, int *width, int *height)
{
	static char *greetings = "Player\n";
	char *response;
	write(sock->sockfd, greetings, strlen(greetings));
	if ((response = sock_read_string_dynamic(sock)) == NULL)
		return -1;
	if (sscanf(response, "%d%d", width, height) != 2) {
		free(response);
		return -1;
	}
	free(response);
	return 0;
}



/***************************************************************************/
/* GUI stuff - excellent for observing robots movements */

WINDOW *pad;

void plot_map(map_t *map)
{
	int x, y, i;
	listiterator_t *it;
	robot_t *robot;
	package_t *pkg;
	char buffer[4096]; 

	wmove(pad, 0, 0);
	//for (y = map->height; y >= 1; y--) {
	for (y = 1; y <= map->height; y++) {
		for (x = 1; x <= map->width; x++) {
			if (map_get_tiletype(map, x, y) == TT_WATER)
				waddch(pad, '~');
			else if (map_get_tiletype(map, x, y) == TT_WALL)
				waddch(pad, '#');
			else if (map_get_tiletype(map, x, y) == TT_PLAIN)
			       	waddch(pad, '.');
			else if (map_get_tiletype(map, x, y) == TT_HOMEBASE) 
				waddch(pad, '@');
			else
				abort();
		}
	}

	for (i = 0; i < packages->size; i++) {
		pkg = package_get(packages->data[i]);
		wattron(pad, COLOR_PAIR(2));
		wmove(pad, pkg->y - 1, pkg->x - 1);
		waddch(pad, 'X');
		wattron(pad, COLOR_PAIR(3));
		wmove(pad, pkg->dy - 1, pkg->dx - 1);
		waddch(pad, '0');
	}

	for (it = robot_iterator(); it->current; it = listiterator_next(it)) {
		robot = listiterator_data(robot_t *, it);
		if (robot->id == myrobot->id)
			wattron(pad, COLOR_PAIR(1));
		else
			wattron(pad, COLOR_PAIR(5));
		wmove(pad, robot->y - 1, robot->x - 1);
		waddch(pad, '0' + (robot->id%10));
	}
	listiterator_free(it);
	wattrset(pad, A_NORMAL);

	wmove(pad, map->height, 0);
	sprintf(buffer, "robot id: %d load: %d", myrobot->id, myrobot->load);
	waddstr(pad, buffer);
	wmove(pad, map->height + 1, 0);
	buffer[0] = '\0';
	for (i = 0; i < myrobot->packages->size; i++)
		sprintf(buffer + strlen(buffer), " %d",
			myrobot->packages->data[i]);	
//	waddstr(pad, buffer);
	
	sprintf(buffer, "%d %d       ", packages->size, myrobot->maxbid);
	waddstr(pad, buffer);

	x = myrobot->x - 1;
	y = myrobot->y - 1;
	x -= COLS / 2;
	y -= LINES / 2;
	if (x < 0)
		x = 0;
	if (y < 0)
		y = 0;
	if (x + COLS >= map->width)
		x = map->width - COLS;
	if (y + LINES >= (map->height + 2))
		y = (map->height + 2) - LINES;
	
	wmove(pad, (map->height + 2) - 1, map->width - 1);
	prefresh(pad, x, y, 0, 0, LINES - 1, COLS - 1);
}

void gui_init(map_t *map)
{
	initscr();
	cbreak();
	noecho();
	nonl();
	start_color();
	init_pair(1, COLOR_WHITE, COLOR_MAGENTA);
	init_pair(2, COLOR_BLACK, COLOR_YELLOW);
	init_pair(3, COLOR_WHITE, COLOR_BLUE);
	init_pair(5, COLOR_WHITE, COLOR_GREEN);
	pad = newpad(map->height + 2, map->width);

}

/***************************************************************************/

void die(void)
{
	delwin(pad);
	endwin();
}

static int wandering = 0;
static size_t destx, desty;

static void set_wandering(void)
{
	random_reachable_location(&destx, &desty);
	wandering = 1;
}

static void move_robot(sockbuf_t *sock, int direction, int bid)
{
	char movestr[256];
	sprintf(movestr, "%d Move ", bid);
	logger("COMMAND: %s ", movestr);
	write(sock->sockfd, movestr, strlen(movestr));
	switch (1 << direction) {
	case MD_NORTH:
		write(sock->sockfd, "N\n", 2);
		logger("N\n");
		break;
	case MD_SOUTH:
		write(sock->sockfd, "S\n", 2);
		logger("S\n");
		break;
	case MD_EAST:
		write(sock->sockfd, "E\n", 2);
		logger("E\n");
		break;
	case MD_WEST:
		write(sock->sockfd, "W\n", 2);
		logger("W\n");
		break;
	default:
		endwin();
		printf("%d\r\n", direction);
		fflush(stdout);
		fputs("logic error 2.\n", stderr);
	}
}

static int carrying_packets = 0;
static int fetching_packets = 0;

void reset_package_distances(intarray_t *pkgs)
{
	int i;
	for (i = 0; i < pkgs->size; i++)
		package_get(pkgs->data[i])->distance = 0;
}

void compute_package_distances(intarray_t *pkgs, int robot_x, int robot_y,
	int destination, map_t *map, int coefficient)
{
	int i, dx, dy;
	path_t *path;

	for (i = 0; i < pkgs->size; i++) {
		package_t *pkg = package_get(pkgs->data[i]);
		if (destination) {
			dx = pkg->dx;
			dy = pkg->dy;
		} else {
			dx = pkg->x;
			dy = pkg->y;
		}
		if ((path = path_compute(map, robot_x, robot_y, 
			dx, dy, 1)) == NULL) 
			pkg->distance += 10000000 * coefficient;
		else {
			pkg->distance += path->goal_distance * coefficient;
			path_free(path);
		}
	}
}

static int package_priority_distance(void *a, void *b)
{
	return ((package_t *) a)->distance - ((package_t *) b)->distance;
}

static int package_sort_func(const void *a, const void *b)
{
	int pa, pb;
	pa = *(int *) a;
	pb = *(int *) b;
	return package_priority_distance(package_get(pa), package_get(pb));
}

void sort_packages_by_distance(intarray_t *pkgs)
{
	qsort(pkgs->data, pkgs->size, sizeof(int), package_sort_func);
}

enum {
	RS_WANDERING		= 0,
	RS_CARRYING_FULL	= 1,
	RS_RET_PACKAGE_SITE	= 2
};

/************************************************************************
 * drop_arrived_packages()
 * 
 * Maybe it is best to drop the packages upon arriving at the drop
 * site, regardless of the current situation.
 */
int drop_arrived_packages(map_t *map, char *cmdbuffer)
{
	int items_to_drop = 0;
	int i;
	cmdbuffer[0] = '\0';
	reset_package_distances(myrobot->packages);
	compute_package_distances(myrobot->packages, 
		myrobot->x, myrobot->y, 1, map, 1);
	/* FIXME: what's the correct bid? */
	strcat(cmdbuffer, "1 Drop");
	for (i = 0; i < myrobot->packages->size; i++) {
		package_t *pkg = package_get(myrobot->packages->data[i]);
		if (pkg->distance != 0)
			continue;
		sprintf(cmdbuffer + strlen(cmdbuffer), " %d", 
			myrobot->packages->data[i]);
		items_to_drop++;
	}
	strcat(cmdbuffer, "\n");
	if (items_to_drop)
		logger("COMMAND: %s\n", cmdbuffer);
	return items_to_drop;
}

/************************************************************************
 * pick_packages()
 */
int pick_packages(map_t *map, char *cmdbuffer)
{
	int items_to_pick = 0;
	int i;
	cmdbuffer[0] = '\0';
	reset_package_distances(packages);
	compute_package_distances(packages, myrobot->x, myrobot->y, 0, map, 1);
	/* FIXME: what's the correct bid? */
	strcat(cmdbuffer, "1 Pick");
	for (i = 0; i < packages->size; i++) {
		package_t *pkg = package_get(packages->data[i]);
		if (pkg->distance != 0)
			continue;
		if (pkg->x == pkg->dx && pkg->y == pkg->dy)
			continue;
		if (intarray_find(myrobot->packages, pkg->id) != -1)
			continue;
		if (myrobot->load + pkg->weight > myrobot->max_load)
			continue;
		sprintf(cmdbuffer + strlen(cmdbuffer), " %d", pkg->id);
		items_to_pick++;
	}
	strcat(cmdbuffer, "\n");
	if (items_to_pick)
		logger("COMMAND: %s\n", cmdbuffer);
	return items_to_pick;
}

int robot_in_location(int x, int y)
{
	listiterator_t *it;
	robot_t *robot;
	for (it = robot_iterator(); it->current; it = listiterator_next(it)) {
		robot = listiterator_data(robot_t *, it);
		if (robot->x == x && robot->y == y) {
			listiterator_free(it);
			return robot->id;
		}
	}
	listiterator_free(it);
	return -1;
}

int move_towards(sockbuf_t *sock, map_t *map, int dx, int dy)
{
	path_t *path;
	int rid, bid;
	robot_t *robot;
	if (myrobot->x == dx && myrobot->y == dy)
		return -1;
	if ((path = path_compute(map, myrobot->x, myrobot->y, 
		dx, dy, 1)) == NULL) 
		return -1;	

	bid = 1;
	if ((rid = robot_in_location(dx, dy)) != -1) 
		bid = myrobot->maxbid;

	move_robot(sock, path->moves[0], bid);
	path_free(path);
	return 0;
}

int package_to_be_fetched = -1;

package_t *random_package(intarray_t *pkgs)
{
	int i;
	i = rand() % pkgs->size;
	return package_get(pkgs->data[i]);
}

int path_distance(map_t *map, int sx, int sy, int dx, int dy)
{
	path_t *path;
	int nsteps;
	if ((path = path_compute(map, sx, sy, dx, dy, 1)) == NULL)
		return 1000000;
	else {
		nsteps = path->nsteps;
		path_free(path);
	}
	return nsteps;
}

int robots_nearer(map_t *map, int x, int y)
{
	listiterator_t *it;
	int md;
	md = path_distance(map, myrobot->x, myrobot->y, x, y);
	for (it = robot_iterator(); it->current; it = listiterator_next(it)) {
		robot_t *robot = listiterator_data(robot_t *, it);
		if (robot->id == myrobot->id)
			continue;
		if (path_distance(map, robot->x, robot->y, x, y) < md + 1) {
			listiterator_free(it);
			return 1;
		}
	}
	listiterator_free(it);
	return 0;
}

int nearest_free_package(robot_t *robot, map_t *map)
{
	int i, free, pkgid;
	robot_t *r;
	listiterator_t *it;
	package_t *pkg;
	reset_package_distances(packages);
	for (it = robot_iterator(); it->current; it = listiterator_next(it)) {
		r = listiterator_data(robot_t *, it);
		if (r->id == myrobot->id)
			compute_package_distances(packages, robot->x, 
				robot->y, 0, map, 1);
		else
			compute_package_distances(packages, robot->x, 
				robot->y, 0, map, -1);
	}
	listiterator_free(it);
	sort_packages_by_distance(packages);
	for (i = 0; i < packages->size; i++) {
		free = 1;
		pkgid = packages->data[i];
		if (map_get_blacklist(map, robot->x, robot->y) > 0) {
			free = 0;
			break;
		}
		for (it = robot_iterator(); it->current && free;
			it = listiterator_next(it)) {
			r = listiterator_data(robot_t *, it);
			if (intarray_find(r->packages, pkgid) != -1) {
				free = 0;
				break;
			}
		}
		listiterator_free(it);
		pkg = package_get(pkgid);		
		if (robots_nearer(map, pkg->x, pkg->y))
			continue;
		if (!free)
			continue;
		return pkgid;
	}
	return -1;
}

int package_nearest_drop_site(map_t *map, int x, int y, intarray_t *pkgs)
{
	listiterator_t *it;
	robot_t *robot;
	int i;
	reset_package_distances(pkgs);
	for (it = robot_iterator(); it->current; it = listiterator_next(it)) {
		robot = listiterator_data(robot_t *, it);
		if (robot->id == myrobot->id) 
			compute_package_distances(pkgs, x, y, 1, map, 1);
		else
			compute_package_distances(pkgs, x, y, 1, map, -1);
	}
	listiterator_free(it);
	sort_packages_by_distance(pkgs);
	return pkgs->data[0];	
}


#define BSZ 0
void blacklist_surroundings(map_t *map, int x, int y)
{
	int tx, ty;
	for (ty = -BSZ; ty <= -BSZ; ty++)
		for (tx = -BSZ; tx <= BSZ; tx++) {
			int bx, by;
			bx = x + tx;
			by = y + ty;
			if (bx < 1 || by < 1 || 
				bx > map->width || by > map->height)
				continue;
			map_increase_blacklist(map, bx, by);
		}
}

void decrease_blacklist(map_t *map)
{
	int x, y;
	for (y = 1; y <= map->height; y++)
		for (x = 1; x <= map->width; x++)
			map_decrease_blacklist(map, x, y);
}

int random_foo(void)
{
	return (rand() % 10) > 7;
}

void remove_nonexistent_packages(void)
{
	int i = 0;
	while (i < packages->size) {
		package_t *pkg = package_get(packages->data[i]);
		if (pkg->x != myrobot->x || pkg->y != myrobot->y) {
			i++;
			continue;
		}
		if (intarray_find(packages_here, pkg->id) == -1)
			intarray_delete(packages, pkg->id);
		else
			i++;
	}
}


int new_idea(sockbuf_t *sock, map_t *map)
{
	path_t *path;
	char cmdbuffer[4096];
	package_t *pkg;

	remove_nonexistent_packages();

	if (wandercount) {
		wandercount--;
		goto begin;
	}

	if (myrobot->pushed || (myrobot->moved && !myrobot->dx && !myrobot->dy)) {
		logger("Robot %d was pushed.\n", myrobot->id);
		if (package_to_be_fetched == -1 &&
			myrobot->packages->size && ((rand() % 10) > 5)) 
			myrobot->maxbid++;
		blacklist_surroundings(map, myrobot->x, myrobot->y);
		wandering = 0;
		wandercount = 3;
		goto begin;
	}
	
	if (drop_arrived_packages(map, cmdbuffer)) {
		write(sock->sockfd, cmdbuffer, strlen(cmdbuffer));
		return 0;
	}
	if (pick_packages(map, cmdbuffer)) {
		write(sock->sockfd, cmdbuffer, strlen(cmdbuffer));
		return 0;
	}

	if (myrobot->packages->size) {
		int pkgid;

		package_to_be_fetched = -1;

		pkgid = package_nearest_drop_site(map, 
			myrobot->x, myrobot->y, myrobot->packages);
		pkg = package_get(pkgid);
		if (move_towards(sock, map, pkg->dx, pkg->dy) == 0) 
			return 0;
	}

	if (packages->size) {
		int pkgid;
	       
		pkgid = nearest_free_package(myrobot, map);
		package_to_be_fetched = pkgid;
	}

	if (package_to_be_fetched != -1) {
		pkg = package_get(package_to_be_fetched);
		if (move_towards(sock, map, pkg->x, pkg->y) == 0) 
			return 0;
	}
		
begin:
	logger("wandering\n");
	if (myrobot->x == destx && myrobot->y == desty) {
		wandering = 0;
	}
	if (!wandering) {
		set_wandering();
		goto begin;
	}
	if (move_towards(sock, map, destx, desty)) {
		wandering = 0;
		goto begin;
	}

	return 0;
}	

int my_atoi(char **s)
{
	int result = 0;
	while (isdigit(**s)) {
		result = result * 10 + **s - '0';
		(*s)++;
	}
	return result;
}

int robot_died(char *buffer)
{
	regex_t reg;
	if (regcomp(&reg, "^Robot [0-9]* died.$", REG_ICASE | REG_NOSUB)) {
		logger("invalid regex\n");
		return -1;
	}
	if (regexec(&reg, buffer, 0, NULL, 0) == 0) {
		while (*buffer && !isdigit(*buffer))
			buffer++;
		regfree(&reg);
		return atoi2(&buffer);
	}
	regfree(&reg);
	return -1;
}	

char *read_server_response(sockbuf_t *sock)
{
	char *response;
	int id;

	if ((response = sock_read_string_dynamic(sock)) == NULL)
		return NULL;

	if ((id = robot_died(response)) != -1) {
		if (myrobot && id != myrobot->id)
			logger("robot %d died\n", id);
		else {	
			endwin();
			printf("Game ended.\n");
			exit(0);
		}
	}

	return response;
}

int process_server_response(sockbuf_t *sock, map_t *map)
{
	char *response, *s;
	int robot_number;
	robot_t *robot;
	static int pc = -2;

	if ((response = read_server_response(sock)) == NULL)
		return -1;
	logger("RESPONSE: %s\n", response);
	
	robot_reset_alive_status();
	
	s = response;
	while (s && *s) {
		while (isspace(*s)) {
			s++;
			continue;
		}
		if (*s == '#') {
			s++;
			robot_number = atoi2(&s);
			logger("processing commands for robot %d\n", robot_number);
			robot = robot_find(robot_number);
			robot->moved = 0;
			robot->dx = 0;
			robot->dy = 0;
			robot->pushed = 0;
			robot->alive = 1;
			continue;
		} else if (*s == 'W') {
			robot_move(robot, map, -1, 0);
			s++;
			continue;
		} else if (*s == 'E') {
			robot_move(robot, map, 1, 0);
			s++;
			continue;
		} else if (*s == 'N') {
			robot_move(robot, map, 0, 1);
			s++;
			continue;
		} else if (*s == 'S') {
			robot_move(robot, map, 0, -1),
			s++;
			continue;
		} else if (*s == 'X') {
			s += 2;
			robot->x = atoi2(&s);
			continue;
		} else if (*s == 'Y') {
			s += 2;
			robot->y = atoi2(&s);
			continue;
		} else if (*s == 'P') {
			package_t *pkg;
			int id, tmp;
			s += 2;
			id = atoi2(&s);

			if (intarray_find(packages, id) == -1) {
				package_new(pc, UNKNOWN, UNKNOWN,
					robot->x, robot->y, UNKNOWN);
				intarray_append(packages, pc);
				pc--;
				package_new(id, UNKNOWN, UNKNOWN,
					robot->x, robot->y, UNKNOWN);
				intarray_append(packages, id);
			}
			robot_add_package(robot, id);

			if (id == package_to_be_fetched &&
				robot->id != myrobot->id)
				package_to_be_fetched = -1;

			package_get(id)->lastpick = iterations;

			continue;
		} else if (*s == 'D') {
			int id;
			s += 2;
			id = atoi2(&s);

			if (package_get(id)->lastpick == iterations)
				robot->pushed = 1;

			robot_del_package(robot, id);
			purge_delivered_packages();

			logger("%d packages left\n", packages->size);
			
			continue;
		}
	}

	free(response);
	return 0;
}			

long int atoi2(char **p)
{
	char *endptr;
	long int result;
	result = strtol(*p, &endptr, 10);
	*p = endptr;
	return result;
}

void purge_dead_robots(void)
{
	listiterator_t *it;
	int i;
	it = robot_iterator();
	while (it->current) {
		robot_t *robot = listiterator_data(robot_t *, it);
		if (!robot->alive) {
			for (i = 0; i < robot->packages->size; i++) 
				intarray_delete(packages, 
					robot->packages->data[i]);
			/* FIXME: memory leak */
			listiterator_delete(it);
		} else
			it = listiterator_next(it);
	}
	listiterator_free(it);
}
			
void purge_delivered_packages(void)
{
	int i = 0;
	while (i < packages->size) {
		package_t *pkg = package_get(packages->data[i]);
		if (pkg->x == pkg->dx && pkg->y == pkg->dy) {
// FIXME:			package_free(pkg);
			intarray_delete(packages, pkg->id);
			logger("package %d purged\n", pkg->id);
		} else
			i++;
	}
}

int process_package_information(sockbuf_t *sock, map_t *map)
{
	char *response, *p;
	int id, dx, dy, weight;

	if ((response = read_server_response(sock)) == NULL)
		return -1;
	
	logger("%d: server informed packages: %s\n", myrobot->id, response);

	intarray_clear(packages_here);

	p = response;
	while (p && *p) {
		id = atoi2(&p);
		dx = atoi2(&p);
		dy = atoi2(&p);
		weight = atoi2(&p);
		intarray_append(packages_here, id);
		if (intarray_find(packages, id) != -1) {
			package_t *pkg = package_get(id);
			pkg->dx = dx;
			pkg->dy = dy;
			pkg->weight = weight;
		} else {
			logger("new package (%d, weight %d) %d,%d -> %d,%d\n", id, 
				weight, myrobot->x, myrobot->y, dx, dy);
			package_new(id, dx, dy, myrobot->x, myrobot->y, weight);
			intarray_append(packages, id);
		}
	}
	free(response);
	return 0;
}

int main(int argc, char *argv[])
{
	map_t *map;
	int i;
	sockbuf_t *sock;
	int board_width, board_height;
	int robot_id, max_load, fuel;

	srand(time(NULL));
//	signal(SIGSEGV, segfault);
	
	if (argc < 3) {
		fputs("koe host port.\n", stderr);
		return 1;
	}
	if ((sock = init_network(argv[1], atoi(argv[2]))) == NULL) {
		fprintf(stderr, "cannot connect %s at port %d.\n", argv[1],
			atoi(argv[2]));
		return 1;
	}
	if (init_game(sock, &board_width, &board_height) == -1) {
		fputs("error initializing game.\n", stderr);
		return 1;
	}
#ifndef NOINFO
	printf("Board size .... : %d x %d\n", board_width, board_height);
#endif
	
	if (robot_init()) {
		fputs("Memory exhausted.\n", stderr);
		return 1;
	}
	init_path_engine();	
	package_init();
	map = map_read(sock, board_width, board_height);

	packages_here = intarray_new();
	packages = intarray_new();

	if ((robot_id = sock_read_number(sock)) == -1 ||
		(max_load = sock_read_number(sock)) == -1 ||
		(fuel = sock_read_number(sock)) == -1) {
		fputs("error initializing game.\n", stderr);
		return 1;
	}

	printf("robot capacity %d\n", max_load);

	process_server_response(sock, map);
	myrobot = robot_find(robot_id);
	myrobot->max_load = max_load;
	printf("my robot's home: %d, %d\n", myrobot->x, myrobot->y);

	compute_reachable_locations(map, myrobot->x, myrobot->y);

	printf("ROBOT INITIALIZATION COMPLETE\n");

	gui_init(map);
	atexit(die);

	for (;;) {
		decrease_blacklist(map);
		compute_price_mask(map, myrobot);
		process_package_information(sock, map);
		new_idea(sock, map);
		process_server_response(sock, map);
		purge_delivered_packages();
		purge_dead_robots();
		plot_map(map);
	}

	return 0;
}


