#ifndef _UTIL_H_INCLUDED
#define _UTIL_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>

typedef struct listlink_tag {
	struct listlink_tag *next;
	void *data;
} listlink_t;

typedef struct {
	listlink_t *list;
	int (*compare)(void *, void *);
#ifdef PROFILE
	int size;
#endif
} list_t;

#define UNKNOWN (1L<<31)

typedef struct {
	listlink_t *prev;
	listlink_t *current;
	list_t *list;
} listiterator_t;

void *list_find(list_t *list, void *key);
void *list_new(int (*compare)(void *, void *));
void list_add(list_t *list, void *data);
void list_add_priority(list_t *list, void *data, int (*pfunc)(void *, void *));
void *list_car_delete(list_t *list);
void list_free(list_t *list);

void *xmalloc(size_t size);
void *xrealloc(void *ptr, size_t size);

static inline void listiterator_free(listiterator_t *it);
static inline listiterator_t *listiterator_next(listiterator_t *it);

static inline listiterator_t *listiterator_new(list_t *list) 
{
	listiterator_t *it;

	if ((it = (listiterator_t *) malloc(sizeof(listiterator_t))) == NULL)
		return NULL;
	it->prev = NULL;
	it->current = list->list;
	it->list = list;
	return it;
}

static inline listiterator_t *listiterator_find(list_t *list, void *key)
{
	listiterator_t *it;
	for (it = listiterator_new(list); it->current; 
		it = listiterator_next(it)) 
		if (list->compare(it->current->data, key))
			break;
	if (it->current == NULL) {
		listiterator_free(it);
		return NULL;
	}
	return it;
}

static inline void listiterator_delete(listiterator_t *it)
{
	listlink_t *link;
	if (!it->current)
		return;
	if (it->prev)
		it->prev->next = it->current->next;
	else
		it->list->list = it->current->next;
	link = it->current->next;
	free(it->current),
	it->current = link;
}

static inline void listiterator_free(listiterator_t *it)
{
	free(it);
}

static inline listiterator_t *listiterator_next(listiterator_t *it)
{
	if (it->current) {
		it->prev = it->current;
		it->current = it->current->next;
	}
	return it;
}

#define listiterator_data(__type__, __it__) \
	((__type__) ((__it__)->current->data))

static inline int list_length(list_t *l)
{
	int size = 0;
	listlink_t *link = l->list;
	while (link) {
		size++;
		link = link->next;
	}
	return size;
}

typedef struct {
	int *data;
	int size;
	int nalloc;
} intarray_t;

static intarray_t *intarray_new(void)
{
	intarray_t *array;
	array = (intarray_t *) malloc(sizeof(intarray_t));
	array->data = (int *) malloc(sizeof(int));
	array->nalloc = 1;
	array->size = 0;
	return array;
}

static void intarray_enlarge(intarray_t *array)
{
	if ((array->size + 1) < array->nalloc)
		return;
	array->nalloc *= 2;
	array->data = (int *) realloc(array->data,
		array->nalloc * sizeof(int));
}

static void intarray_append(intarray_t *array, int i)
{
	intarray_enlarge(array);
	array->data[array->size++] = i;
}

static void intarray_clear(intarray_t *array)
{
	array->size = 0;
}

static int intarray_find(intarray_t *array, int i)
{
	int j;
	for (j = 0; j < array->size; j++)
		if (array->data[j] == i)
			return j;
	return -1;
}

static int intarray_delete(intarray_t *array, int i)
{
	int pos;
       	pos = intarray_find(array, i);
	if (pos == -1)
		return -1;
	for (; pos < array->size - 1; pos++)
		array->data[pos] = array->data[pos + 1];
	array->size--;
	return 0;
}

#ifdef __cplusplus
};
#endif

#endif /* _UTIL_H_INCLUDED */
